@@ -1,22 +1,36 @@
 """Dataset class for defining data source, splitting, parsing, and iteration."""
 
 import json
+from dataclasses import _MISSING_TYPE, field, make_dataclass
+from enum import Enum
 from functools import partial
-from inspect import Parameter, signature
+from inspect import Parameter, _empty, signature
 from pathlib import Path
-from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Type, TypeVar, cast
+from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Tuple, Type, TypeVar, Union, cast
+
+try:
+    from typing import get_args  # type: ignore
+except ImportError:
+    from typing_extensions import get_args
 
 import pandas as pd
+from dataclasses_json import dataclass_json
 from flytekit.core.tracker import TrackedInstance
 from flytekit.extras.sqlite3.task import SQLite3Task
 from sklearn.model_selection import train_test_split
 
+import unionml.type_guards as type_guards
 from unionml.utils import inner_task
 
 R = TypeVar("R")  # raw data
 D = TypeVar("D")  # model-ready data
 
 
+class ReaderReturnTypeSource(Enum):
+    READER = "reader"
+    LOADER = "loader"
+
+
 class Dataset(TrackedInstance):
     def __init__(
         self,
@@ -68,9 +82,14 @@ class Dataset(TrackedInstance):
 
         self._reader = None
         self._reader_input_types: Optional[List[Parameter]] = None
-        self._reader_return_type: Optional[Dict[str, Type]] = None
+        self._dataset_datatype: Optional[Dict[str, Type]] = None
         self._dataset_task = None
 
+        # dynamically defined types
+        self._loader_kwargs_type: Optional[Type] = None
+        self._splitter_kwargs_type: Optional[Type] = None
+        self._parser_kwargs_type: Optional[Type] = None
+
     def reader(self, fn=None, **reader_task_kwargs):
         """Register a reader function for getting data from some external source.
 
@@ -80,6 +99,8 @@ class Dataset(TrackedInstance):
         """
         if fn is None:
             return partial(self.reader, **reader_task_kwargs)
+
+        type_guards.guard_reader(fn)
         self._reader = fn
         self._reader_task_kwargs = reader_task_kwargs
         return fn
@@ -95,6 +116,7 @@ class Dataset(TrackedInstance):
 
         :param fn: function to register
         """
+        type_guards.guard_loader(fn, self.dataset_datatype["data"])
         self._loader = fn
         return fn
 
@@ -118,6 +140,8 @@ class Dataset(TrackedInstance):
                 n = int(data.shape[0] * test_size)
                 return data.iloc[:-n], data.iloc[-n:]
         """
+
+        type_guards.guard_splitter(fn, self.dataset_datatype["data"], self.dataset_datatype_source.value)
         self._splitter = fn
         return fn
 
@@ -142,6 +166,7 @@ class Dataset(TrackedInstance):
                     features = [col for col in data if col not in targets]
                 return data[features], data[targets]
         """
+        type_guards.guard_parser(fn, self.dataset_datatype["data"], self.dataset_datatype_source.value)
         self._parser = fn
         self._parser_feature_key = feature_key
         return fn
@@ -156,6 +181,7 @@ class Dataset(TrackedInstance):
 
         And it should return the data structure needed for model training.
         """
+        type_guards.guard_feature_loader(fn, self.parser_return_types[self._parser_feature_key])
         self._feature_loader = fn
         return fn
 
@@ -168,6 +194,7 @@ class Dataset(TrackedInstance):
         2. When the FastAPI app `/predict/` endpoint is invoked with features passed in as JSON or string encoding.
         3. When the `model.predict` or `model.remote_predict` functions are invoked.
         """
+        type_guards.guard_feature_transformer(fn, self.parser_return_types[self._parser_feature_key])
         self._feature_transformer = fn
         return fn
 
@@ -188,6 +215,57 @@ class Dataset(TrackedInstance):
             "targets": self._targets,
         }
 
+    @staticmethod
+    def _kwarg_to_dataclass_field(param, default_kwargs):
+        default = default_kwargs.get(param.name, _MISSING_TYPE if param.default is _empty else param.default)
+        if isinstance(default, (list, dict, set)):
+            _kwargs = {"default_factory": lambda: default}
+        else:
+            _kwargs = {"default": default}
+        return field(**_kwargs)
+
+    @property
+    def loader_kwargs_type(self):
+        """Dynamically create a dataclass for loader kwargs."""
+        if self._loader_kwargs_type is not None:
+            return self._loader_kwargs_type
+
+        fields = [
+            (p.name, p.annotation, field() if p.default is _empty else field(default=p.default))
+            for i, p in enumerate(signature(self._loader).parameters.values())
+            if i > 0  # skip the first argument, which is the raw data
+        ]
+        self._loader_kwargs_type = dataclass_json(make_dataclass("LoaderKwargs", fields))
+        return self._loader_kwargs_type
+
+    @property
+    def splitter_kwargs_type(self):
+        """Dynamically create a dataclass for splitter kwargs."""
+        if self._splitter_kwargs_type is not None:
+            return self._splitter_kwargs_type
+
+        fields = [
+            (p.name, p.annotation, self._kwarg_to_dataclass_field(p, self.splitter_kwargs))
+            for i, p in enumerate(signature(self._splitter).parameters.values())
+            if i > 0  # skip the first argument, which is the data
+        ]
+        self._splitter_kwargs_type = dataclass_json(make_dataclass("SplitterKwargs", fields))
+        return self._splitter_kwargs_type
+
+    @property
+    def parser_kwargs_type(self):
+        """Dynamically create a dataclass for parser kwargs."""
+        if self._parser_kwargs_type is not None:
+            return self._parser_kwargs_type
+
+        fields = [
+            (p.name, p.annotation, self._kwarg_to_dataclass_field(p, self.parser_kwargs))
+            for i, p in enumerate(signature(self._parser).parameters.values())
+            if i > 0  # skip the first argument, which is the data
+        ]
+        self._parser_kwargs_type = dataclass_json(make_dataclass("ParserKwargs", fields))
+        return self._parser_kwargs_type
+
     def dataset_task(self):
         """Create a Flyte task for getting the dataset using the ``reader`` function."""
         if self._dataset_task:
@@ -208,7 +286,13 @@ class Dataset(TrackedInstance):
         self._dataset_task = dataset_task
         return dataset_task
 
-    def get_data(self, raw_data):
+    def get_data(
+        self,
+        raw_data,
+        loader_kwargs: Optional[Dict[str, Any]] = None,
+        splitter_kwargs: Optional[Dict[str, Any]] = None,
+        parser_kwargs: Optional[Dict[str, Any]] = None,
+    ) -> Dict[str, Any]:
         """Get training data from from its raw form to its model-ready form.
 
         :param raw_data: Raw data in the same form as the ``reader`` output.
@@ -219,15 +303,26 @@ class Dataset(TrackedInstance):
         - :meth:`unionml.dataset.Dataset.splitter`
         - :meth:`unionml.dataset.Dataset.parser`
         """
-        data = self._loader(raw_data)
-        splits = self._splitter(data, **self.splitter_kwargs)
+
+        def update_kwargs(
+            kwargs: Dict[str, Any],
+            new_kwargs: Optional[Dict[str, Any]],
+        ) -> Dict[str, Any]:
+            return kwargs if new_kwargs is None else {**kwargs, **new_kwargs}
+
+        loader_kwargs = update_kwargs({}, loader_kwargs)
+        splitter_kwargs = update_kwargs(self.splitter_kwargs, splitter_kwargs)
+        parser_kwargs = update_kwargs(self.parser_kwargs, parser_kwargs)
+
+        data = self._loader(raw_data, **loader_kwargs)
+        splits = self._splitter(data, **splitter_kwargs)
+
         if len(splits) == 1:
-            return {"train": self._parser(splits[0], **self.parser_kwargs)}
+            return {"train": self._parser(splits[0], **parser_kwargs)}
 
         train_split, test_split = splits
-        # TODO: make this more generic so as to include a validation split
-        train_data = self._parser(train_split, **self.parser_kwargs)
-        test_data = self._parser(test_split, **self.parser_kwargs)
+        train_data = self._parser(train_split, **parser_kwargs)
+        test_data = self._parser(test_split, **parser_kwargs)
         return {
             "train": train_data,
             "test": test_data,
@@ -242,8 +337,9 @@ class Dataset(TrackedInstance):
         - :meth:`unionml.dataset.Dataset.parser`
         - :meth:`unionml.dataset.Dataset.feature_transformer`
         """
-        parsed_data = self._parser(self._feature_loader(features), self._features, self._targets)
-        features = parsed_data[self._parser_feature_key]
+        # parsed_data = self._parser(self._feature_loader(features), self._features, self._targets)
+        # features = parsed_data[self._parser_feature_key]
+        features = self._feature_loader(features)
         return self._feature_transformer(features)
 
     @property
@@ -254,21 +350,53 @@ class Dataset(TrackedInstance):
         return self._reader_input_types
 
     @property
-    def reader_return_type(self) -> Dict[str, Type]:
-        """Get the output type of the ``reader`` .
+    def dataset_datatype(self) -> Dict[str, Type]:
+        """Get the output type of the ``reader`` or a user-defined ``loader``.
 
-        If the ``loader`` is user-defined then the output of that function is used.
+        The type from the ``loader`` takes precedence.
         """
         if self._loader != self._default_loader:
             return {"data": signature(self._loader).return_annotation}
-        if self._reader and self._reader_return_type is None:
+        if self._reader and self._dataset_datatype is None:
             return {"data": signature(self._reader).return_annotation}
-        elif self._reader_return_type is not None:
-            return self._reader_return_type
+        elif self._dataset_datatype is not None:
+            return self._dataset_datatype
         raise ValueError(
-            "reader_return_type is not defined. Please define a @dataset.reader function with an output annotation."
+            "dataset_datatype is not defined. Please define a @dataset.reader function with an output annotation."
         )
 
+    @property
+    def dataset_datatype_source(self) -> ReaderReturnTypeSource:
+        """Get the output type of the ``reader`` or a user-defined ``loader``.
+
+        The type from the ``loader`` takes precedence.
+        """
+        return ReaderReturnTypeSource.LOADER if self._loader != self._default_loader else ReaderReturnTypeSource.READER
+
+    @property
+    def parser_return_types(self) -> Iterable[Type]:
+        """Get an iterable of types produces by the parser."""
+        sig = signature(self._parser)
+        return get_args(sig.return_annotation)
+
+    @property
+    def feature_type(self) -> Type:
+        """Get the type returned by the ``feature_transformer``, falling back to the ``parser``.
+
+        The fallback behavior occurs if the user didn't define a ``feature_transformer`` function.
+        """
+        parser_type = get_args(signature(self._parser).return_annotation)[self._parser_feature_key]
+
+        ft_type = signature(self._feature_transformer).return_annotation
+        if self._feature_transformer == self._default_feature_transformer:
+            # if the feature transformer is not user-defined, use the parser return signature
+            ft_type = get_args(signature(self._parser).return_annotation)[self._parser_feature_key]
+
+        if parser_type != ft_type:
+            return cast(Type, Union[ft_type, parser_type])
+
+        return parser_type
+
     @classmethod
     def _from_flytekit_task(
         cls,
@@ -278,7 +406,7 @@ class Dataset(TrackedInstance):
     ) -> "Dataset":
         dataset = cls(*args, **kwargs)
         dataset._dataset_task = task
-        dataset._reader_return_type = task.python_interface.outputs
+        dataset._dataset_datatype = task.python_interface.outputs
         dataset._reader_input_types = [
             Parameter(k, Parameter.KEYWORD_ONLY, annotation=v) for k, v in task.python_interface.inputs.items()
         ]
@@ -315,7 +443,7 @@ class Dataset(TrackedInstance):
         return cls._from_flytekit_task(task, *args, **kwargs)
 
     def _default_loader(self, data: R) -> R:
-        [(_, data_type)] = self.reader_return_type.items()
+        [(_, data_type)] = self.dataset_datatype.items()
         if data_type is pd.DataFrame:
             return pd.DataFrame(data)
         return data
@@ -331,9 +459,14 @@ class Dataset(TrackedInstance):
             return (data,)
         return train_test_split(data, test_size=test_size, random_state=random_state, shuffle=shuffle)
 
-    def _default_parser(self, data: D, features: Optional[List[str]], targets: Optional[List[str]]) -> Tuple[D, ...]:
+    def _default_parser(
+        self,
+        data: pd.DataFrame,
+        features: Optional[List[str]],
+        targets: Optional[List[str]],
+    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
         if not isinstance(data, pd.DataFrame):
-            return (data,)
+            return (data,)  # type: ignore
 
         if features is not None and targets is not None:
             features = [col for col in data if col not in targets]
@@ -348,9 +481,15 @@ class Dataset(TrackedInstance):
             with features.open() as f:
                 features = json.load(f)
 
-        [(_, data_type)] = self.reader_return_type.items()
+        [(_, data_type)] = self.dataset_datatype.items()
+
         if data_type is pd.DataFrame:
-            return pd.DataFrame(features)
+            data = pd.DataFrame(features)
+            features_names = self._features
+            if not features_names and self._targets is not None:
+                features_names = [col for col in data if col not in self._targets]
+            return data[features_names]
+
         return features
 
     def _default_feature_transformer(self, features: R) -> D:
