@@ -18,6 +18,7 @@ from flytekit.core.tracker import TrackedInstance
 from flytekit.remote import FlyteRemote
 from flytekit.remote.executions import FlyteWorkflowExecution
 
+import unionml.type_guards as type_guards
 from unionml.dataset import Dataset
 from unionml.utils import inner_task, is_keras_model, is_pytorch_model
 
@@ -186,6 +187,8 @@ class Model(TrackedInstance):
         """Register a function for training a model object."""
         if fn is None:
             return partial(self.trainer, **train_task_kwargs)
+
+        type_guards.guard_trainer(fn, self.model_type, self._dataset.parser_return_types)
         self._trainer = fn
         self._train_task_kwargs = train_task_kwargs
         return self._trainer
@@ -194,12 +197,15 @@ class Model(TrackedInstance):
         """Register a function that generates predictions from a model object."""
         if fn is None:
             return partial(self.predictor, **predict_task_kwargs)
+
+        type_guards.guard_predictor(fn, self.model_type, self._dataset.feature_type)
         self._predictor = fn
         self._predict_task_kwargs = predict_task_kwargs
         return self._predictor
 
     def evaluator(self, fn):
         """Register a function for producing metrics for given model object."""
+        type_guards.guard_evaluator(fn, self.model_type, self._dataset.parser_return_types)
         self._evaluator = fn
         return self._evaluator
 
@@ -237,11 +243,16 @@ class Model(TrackedInstance):
         # add hyperparameter argument
         wf.add_workflow_input(hyperparam_arg, hyperparam_type)
 
+        # add loader, splitter, parser kwargs
+        wf.add_workflow_input("loader_kwargs", self._dataset.loader_kwargs_type)
+        wf.add_workflow_input("splitter_kwargs", self._dataset.splitter_kwargs_type)
+        wf.add_workflow_input("parser_kwargs", self._dataset.parser_kwargs_type)
+
         # add dataset.reader arguments
         for arg, type in dataset_task.python_interface.inputs.items():
             wf.add_workflow_input(arg, type)
 
-        # add training keyword-only arguments
+        # add training kwargs
         trainer_param_types = {k: v.annotation for k, v in self.trainer_params.items()}
         for arg, type in trainer_param_types.items():
             wf.add_workflow_input(arg, type)
@@ -256,6 +267,7 @@ class Model(TrackedInstance):
                 hyperparam_arg: wf.inputs[hyperparam_arg],
                 **dataset_node.outputs,
                 **{arg: wf.inputs[arg] for arg in trainer_param_types},
+                **{arg: wf.inputs[arg] for arg in ["loader_kwargs", "splitter_kwargs", "parser_kwargs"]},
             },
         )
         wf.add_workflow_output("model_object", train_node.outputs["model_object"])
@@ -311,8 +323,8 @@ class Model(TrackedInstance):
         *_, hyperparameters_param = signature(self._init).parameters.values()
         hyperparameters_param = hyperparameters_param.replace(annotation=self.hyperparameter_type)
 
-        # assume that reader_return_type is a dict with only a single entry
-        [(data_arg_name, data_arg_type)] = self._dataset.reader_return_type.items()
+        # assume that dataset_datatype is a dict with only a single entry
+        [(data_arg_name, data_arg_type)] = self._dataset.dataset_datatype.items()
 
         # get keyword-only training args
         @inner_task(
@@ -321,12 +333,20 @@ class Model(TrackedInstance):
                 [
                     (p.name, p)
                     for p in [
+                        # hyperparameters
                         hyperparameters_param,
+                        # raw data
                         Parameter(
                             data_arg_name,
                             kind=Parameter.KEYWORD_ONLY,
                             annotation=data_arg_type,
                         ),
+                        # loader, splitter, and parser kwargs
+                        *[
+                            Parameter(arg, kind=Parameter.KEYWORD_ONLY, annotation=dict)
+                            for arg in ["loader_kwargs", "splitter_kwargs", "parser_kwargs"]
+                        ],
+                        # trainer kwargs
                         *self.trainer_params.values(),
                     ]
                 ]
@@ -371,8 +391,8 @@ class Model(TrackedInstance):
         model_param, *_ = predictor_sig.parameters.values()
         model_param = model_param.replace(name="model_object")
 
-        # assume that reader_return_type is a dict with only a single entry
-        [(data_arg_name, data_arg_type)] = self._dataset.reader_return_type.items()
+        # assume that dataset_datatype is a dict with only a single entry
+        [(data_arg_name, data_arg_type)] = self._dataset.dataset_datatype.items()
         data_param = Parameter(data_arg_name, kind=Parameter.KEYWORD_ONLY, annotation=data_arg_type)
 
         # TODO: make sure return type is not None
@@ -402,8 +422,8 @@ class Model(TrackedInstance):
         model_param, *_ = predictor_sig.parameters.values()
         model_param = model_param.replace(name="model_object")
 
-        # assume that reader_return_type is a dict with only a single entry
-        [(_, data_arg_type)] = self._dataset.reader_return_type.items()
+        # assume that dataset_datatype is a dict with only a single entry
+        [(_, data_arg_type)] = self._dataset.dataset_datatype.items()
         data_param = Parameter("features", kind=Parameter.KEYWORD_ONLY, annotation=data_arg_type)
 
         @inner_task(
@@ -421,6 +441,9 @@ class Model(TrackedInstance):
     def train(
         self,
         hyperparameters: Optional[Dict[str, Any]] = None,
+        loader_kwargs: Optional[Dict[str, Any]] = None,
+        splitter_kwargs: Optional[Dict[str, Any]] = None,
+        parser_kwargs: Optional[Dict[str, Any]] = None,
         trainer_kwargs: Optional[Dict[str, Any]] = None,
         **reader_kwargs,
     ) -> Tuple[Any, Any]:
@@ -428,6 +451,12 @@ class Model(TrackedInstance):
 
         :param hyperparameters: a dictionary mapping hyperparameter names to values. This is passed into the
             ``init`` callable to initialize a model object.
+        :param loader_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.loader` function.
+            This will override any defaults set in the function definition.
+        :param splitter_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.splitter` function.
+            This will override any defaults set in the function definition.
+        :param parser_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.parser` function.
+            This will override any defaults set in the function definition.
         :param trainer_kwargs: a dictionary mapping training parameter names to values. There training parameters
             are determined by the keyword-only arguments of the ``model.trainer`` function.
         :param reader_kwargs: keyword arguments that correspond to the :meth:`unionml.Dataset.reader` method signature.
@@ -435,6 +464,9 @@ class Model(TrackedInstance):
         trainer_kwargs = {} if trainer_kwargs is None else trainer_kwargs
         model_obj, hyperparameters, metrics = self.train_workflow()(
             hyperparameters=self.hyperparameter_type(**({} if hyperparameters is None else hyperparameters)),
+            loader_kwargs=self._dataset.loader_kwargs_type(**({} if loader_kwargs is None else loader_kwargs)),
+            splitter_kwargs=self._dataset.splitter_kwargs_type(**({} if splitter_kwargs is None else splitter_kwargs)),
+            parser_kwargs=self._dataset.parser_kwargs_type(**({} if parser_kwargs is None else parser_kwargs)),
             **{**reader_kwargs, **trainer_kwargs},
         )
         self.artifact = ModelArtifact(model_obj, hyperparameters, metrics)
@@ -454,7 +486,6 @@ class Model(TrackedInstance):
             following order:
 
             - :meth:`unionml.dataset.Dataset.feature_loader`
-            - :meth:`unionml.dataset.Dataset.parser`
             - :meth:`unionml.dataset.Dataset.feature_transformer`
         :param reader_kwargs: keyword arguments that correspond to the :meth:`unionml.Dataset.reader` method signature.
         """
@@ -533,11 +564,14 @@ class Model(TrackedInstance):
         )
         return self.__remote__
 
-    def remote_deploy(self):
+    def remote_deploy(self, app_version: str = None):
         """Deploy model services to a Flyte backend."""
+        if self._remote is None:
+            raise RuntimeError("First configure the remote client with the `Model.remote` method")
+
         from unionml import remote
 
-        app_version = remote.get_app_version()
+        app_version = app_version or remote.get_app_version(allow_uncommitted=True)
         image = remote.get_image_fqn(self, app_version, self._image_name)
 
         os.environ["FLYTE_INTERNAL_IMAGE"] = image or ""
@@ -564,6 +598,9 @@ class Model(TrackedInstance):
         wait: bool = True,
         *,
         hyperparameters: Optional[Dict[str, Any]] = None,
+        loader_kwargs: Optional[Dict[str, Any]] = None,
+        splitter_kwargs: Optional[Dict[str, Any]] = None,
+        parser_kwargs: Optional[Dict[str, Any]] = None,
         trainer_kwargs: Optional[Dict[str, Any]] = None,
         **reader_kwargs,
     ) -> Union[ModelArtifact, FlyteWorkflowExecution]:
@@ -575,6 +612,12 @@ class Model(TrackedInstance):
             function returns a ``FlyteWorkflowExecution``.
         :param hyperparameters: a dictionary mapping hyperparameter names to values. This is passed into the
             ``init`` callable to initialize a model object.
+        :param loader_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.loader` function.
+            This will override any defaults set in the function definition.
+        :param splitter_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.splitter` function.
+            This will override any defaults set in the function definition.
+        :param parser_kwargs: key-word arguments to pass to the registered :class:`unionml.Dataset.parser` function.
+            This will override any defaults set in the function definition.
         :param trainer_kwargs: a dictionary mapping training parameter names to values. There training parameters
             are determined by the keyword-only arguments of the ``model.trainer`` function.
         :param reader_kwargs: keyword arguments that correspond to the :meth:`unionml.Dataset.reader` method signature.
@@ -584,18 +627,26 @@ class Model(TrackedInstance):
 
         from unionml import remote
 
-        app_version = app_version or remote.get_app_version()
+        app_version = app_version or remote.get_app_version(allow_uncommitted=True)
         train_wf = self._remote.fetch_workflow(name=self.train_workflow_name, version=app_version)
         execution = self._remote.execute(
             train_wf,
             inputs={
                 "hyperparameters": self.hyperparameter_type(**({} if hyperparameters is None else hyperparameters)),
+                "loader_kwargs": self._dataset.loader_kwargs_type(**loader_kwargs or {}),
+                "splitter_kwargs": self._dataset.splitter_kwargs_type(**splitter_kwargs or {}),
+                "parser_kwargs": self._dataset.parser_kwargs_type(**parser_kwargs or {}),
                 **{**reader_kwargs, **({} if trainer_kwargs is None else trainer_kwargs)},  # type: ignore
             },
             project=self._remote.default_project,
             domain=self._remote.default_domain,
             wait=wait,
-            type_hints={"hyperparameters": self.hyperparameter_type},
+            type_hints={
+                "hyperparameters": self.hyperparameter_type,
+                "loader_kwargs": self._dataset.loader_kwargs_type,
+                "splitter_kwargs": self._dataset.splitter_kwargs_type,
+                "parser_kwargs": self._dataset.parser_kwargs_type,
+            },
         )
         console_url = self._remote.generate_console_url(execution)
         print(
@@ -632,7 +683,6 @@ class Model(TrackedInstance):
             following order:
 
             - :meth:`unionml.dataset.Dataset.feature_loader`
-            - :meth:`unionml.dataset.Dataset.parser`
             - :meth:`unionml.dataset.Dataset.feature_transformer`
         :param reader_kwargs: keyword arguments that correspond to the :meth:`unionml.Dataset.reader` method signature.
         """
@@ -641,7 +691,7 @@ class Model(TrackedInstance):
 
         from unionml import remote
 
-        app_version = app_version or remote.get_app_version()
+        app_version = app_version or remote.get_app_version(allow_uncommitted=True)
         model_artifact = remote.get_model_artifact(self, app_version, model_version)
 
         if (features is not None and len(reader_kwargs) > 0) or (features is None and len(reader_kwargs) == 0):
@@ -655,7 +705,7 @@ class Model(TrackedInstance):
         else:
             workflow_name = self.predict_from_features_workflow_name
             inputs.update({"features": self._dataset.get_features(features)})
-            type_hints = {"features": [*self._dataset.reader_return_type.values()][0]}
+            type_hints = {"features": [*self._dataset.dataset_datatype.values()][0]}
 
         predict_wf = self._remote.fetch_workflow(
             self._remote._default_project,
@@ -715,7 +765,7 @@ class Model(TrackedInstance):
         """
         from unionml import remote
 
-        app_version = app_version or remote.get_app_version()
+        app_version = app_version or remote.get_app_version(allow_uncommitted=True)
         return remote.list_model_versions(self, app_version=app_version, limit=limit)
 
     def remote_fetch_predictions(self, execution: FlyteWorkflowExecution) -> Any:
@@ -729,6 +779,11 @@ class Model(TrackedInstance):
         predictions, *_ = execution.outputs.values()
         return predictions
 
+    @property
+    def model_type(self) -> Type:
+        init = self._init_callable if self._init == self._default_init else self._init or self._init_callable
+        return init if inspect.isclass(init) else signature(init).return_annotation if init is not None else init
+
     def _default_init(self, hyperparameters: dict) -> Any:
         if self._init_callable is None:
             raise ValueError(
@@ -744,8 +799,7 @@ class Model(TrackedInstance):
         *args,
         **kwargs,
     ) -> Any:
-        init = self._init_callable if self._init == self._default_init else self._init or self._init_callable
-        model_type = init if inspect.isclass(init) else signature(init).return_annotation if init is not None else init
+        model_type = self.model_type
         hyperparameters = (
             asdict(hyperparameters)
             if hyperparameters is not None and is_dataclass(hyperparameters)
@@ -772,9 +826,7 @@ class Model(TrackedInstance):
         )
 
     def _default_loader(self, file: Union[str, os.PathLike, IO], *args, **kwargs) -> Any:
-        init = self._init_callable if self._init == self._default_init else self._init or self._init_callable
-        model_type = init if inspect.isclass(init) else signature(init).return_annotation if init is not None else init
-
+        model_type = self.model_type
         if issubclass(model_type, sklearn.base.BaseEstimator):
             deserialized_model = joblib.load(file, *args, **kwargs)
             return deserialized_model["model_obj"]
