@@ -1,5 +1,6 @@
 """Module for flyte remote helper functions."""
 
+import contextlib
 import importlib
 import logging
 import typing
@@ -40,19 +41,34 @@ def create_project(remote: FlyteRemote, project: typing.Optional[str]):
         remote.client.register_project(Project(id=project, name=project, description=project))
 
 
-def get_app_version() -> str:
+def get_app_version(allow_uncommitted: bool = False) -> str:
     repo = git.Repo(".", search_parent_directories=True)
-    commit = repo.rev_parse("HEAD")
-    return commit.hexsha
+    if repo.is_dirty():
+        if not allow_uncommitted:
+            raise RuntimeError("Version number cannot be determined with uncommitted changes present.")
+        logger.warning("You have uncommitted changes, unionml is using the the latest commit as the app version.")
+
+    with contextlib.suppress(git.CommandError):
+        if list(repo.iter_commits("@{upstream}..")):
+            logger.warning(
+                "You have local commits that are not present on remote repositories,"
+                " which may cause issues with deployment."
+            )
+
+    return repo.head.commit.hexsha
 
 
-def get_image_fqn(model: Model, version: str, image_name: typing.Optional[str] = None) -> str:
+def get_image_fqn(model: Model, app_version: str, image_name: typing.Optional[str] = None) -> str:
     image_name = IMAGE_NAME if image_name is None else image_name
-    return f"{model.registry}/{image_name}:{model.name.replace('_', '-')}-{version}"
+    if model.registry is None:
+        image_uri = image_name
+    else:
+        image_uri = f"{model.registry}/{image_name}"
+    return f"{image_uri}:{model.name.replace('_', '-')}-{app_version}"
 
 
 def sandbox_docker_build(model: Model, image_fqn: str):
-    logger.info("Building docker container in flyte sandbox.")
+    logger.info("Building docker container in flyte demo cluster.")
     client = docker.from_env()
     sandbox_container = None
     for container in client.containers.list():
@@ -61,33 +77,34 @@ def sandbox_docker_build(model: Model, image_fqn: str):
 
     if sandbox_container is None:
         raise RuntimeError(
-            "Cannot find Flyte Sandbox. Make sure to install flytectl and create a sandbox with "
-            "`flytectl sandbox start --source .`"
+            "Cannot find Flyte Demo Cluster. Make sure to install flytectl and create a sandbox with "
+            "`flytectl demo start --source .`"
         )
 
-    _, build_logs = sandbox_container.exec_run(["docker", "build", "/root", "--tag", image_fqn])
-    for line in build_logs.decode().splitlines():
-        logger.info(line)
+    _, build_logs = sandbox_container.exec_run(["docker", "build", "/root", "--tag", image_fqn], stream=True)
+    for line in build_logs:
+        logger.info(line.decode().strip())
 
 
 def docker_build_push(model: Model, image_fqn: str) -> docker.models.images.Image:
+    if model.registry is None:
+        raise ValueError("You must specify a registry in `model.remote` when deploying to a remote cluster.")
+
     client = docker.from_env()
 
     logger.info(f"Building image: {image_fqn}")
-    image, build_logs = client.images.build(
+    build_logs = client.api.build(
         path=".",
         dockerfile=model.dockerfile,
         tag=image_fqn,
         rm=True,
     )
     for line in build_logs:
-        logger.info(line)
+        logger.info(line.decode().strip())
 
-    for line in client.api.push(image.tags[0], stream=True, decode=True):
+    for line in client.api.push(image_fqn, stream=True, decode=True):
         logger.info(line)
 
-    return image
-
 
 def deploy_wf(wf, remote: FlyteRemote, image: str, project: str, domain: str, version: str):
     """Register all tasks, workflows, and launchplans needed to execute the workflow."""
@@ -100,31 +117,73 @@ def deploy_wf(wf, remote: FlyteRemote, image: str, project: str, domain: str, ve
     remote.register_workflow(wf, serialization_settings, version)
 
 
-def get_latest_model_artifact(model: Model, app_version: typing.Optional[str] = None) -> ModelArtifact:
+def get_model_execution(
+    model: Model,
+    app_version: typing.Optional[str] = None,
+    model_version: typing.Optional[str] = "latest",
+) -> FlyteWorkflowExecution:
     if model._remote is None:
         raise RuntimeError("You need to configure the remote client with the `Model.remote` method")
 
-    app_version = get_app_version()
+    app_version = app_version or get_app_version(allow_uncommitted=True)
     train_wf = model._remote.fetch_workflow(
         model._remote._default_project,
         model._remote._default_domain,
         model.train_workflow_name,
         app_version,
     )
-    [latest_training_execution, *_], _ = model._remote.client.list_executions_paginated(
+    if model_version is not None and model_version != "latest":
+        execution = model._remote.fetch_execution(
+            project=train_wf.id.project,
+            domain=train_wf.id.domain,
+            name=model_version,
+        )
+    else:
+        [execution, *_], _ = model._remote.client.list_executions_paginated(
+            train_wf.id.project,
+            train_wf.id.domain,
+            limit=1,
+            filters=[
+                filters.Equal("launch_plan.name", train_wf.id.name),
+                filters.Equal("phase", "SUCCEEDED"),
+            ],
+            sort_by=Sort("created_at", Sort.Direction.DESCENDING),
+        )
+        execution = FlyteWorkflowExecution.promote_from_model(execution)
+    model._remote.sync(execution)
+    return execution
+
+
+def get_model_artifact(
+    model: Model,
+    app_version: typing.Optional[str] = None,
+    model_version: typing.Optional[str] = "latest",
+) -> ModelArtifact:
+    execution = get_model_execution(model, app_version, model_version)
+    model.remote_load(execution)
+    assert model.artifact is not None
+    return model.artifact
+
+
+def list_model_versions(model: Model, app_version: typing.Optional[str] = None, limit: int = 10) -> typing.List[str]:
+    if model._remote is None:
+        raise RuntimeError("You need to configure the remote client with the `Model.remote` method")
+
+    app_version = app_version or get_app_version(allow_uncommitted=True)
+    train_wf = model._remote.fetch_workflow(
+        model._remote._default_project,
+        model._remote._default_domain,
+        model.train_workflow_name,
+        app_version,
+    )
+    executions, _ = model._remote.client.list_executions_paginated(
         train_wf.id.project,
         train_wf.id.domain,
-        limit=1,
+        limit=limit,
         filters=[
             filters.Equal("launch_plan.name", train_wf.id.name),
             filters.Equal("phase", "SUCCEEDED"),
         ],
         sort_by=Sort("created_at", Sort.Direction.DESCENDING),
     )
-    latest_training_execution = FlyteWorkflowExecution.promote_from_model(latest_training_execution)
-    model._remote.sync(latest_training_execution)
-
-    return ModelArtifact(
-        latest_training_execution.outputs["trained_model"],
-        latest_training_execution.outputs["metrics"],
-    )
+    return [x.id.name for x in executions]
